#######################################################################
#Simple Spring System by Jakub Kupikowski                             #
#                                                                     #
#to do:                                                               #
#-add right-click k edit                                              #
#-check speed of attributes vs arays of parameters and implications   #
#######################################################################

require 'sketchup'

#METHODS

#makes vertices array 
def list_vertices(entities)
 list = [] 
 entities.each do |entity|
  if entity.kind_of?(Sketchup::Edge)
   list.push entity.vertices[0] 
   list.push entity.vertices[1]
  end
 end
 list=list.uniq 
end

#makes array of edges
def list_edges(entities)
 list = [] 
 entities.each do |entity|
  if entity.kind_of?(Sketchup::Edge)
   list.push entity
  end
 end
 list
end

#writes initial length to edge's attribute
def remember_length(edge)
 if (edge.get_attribute "info", "initial_length").eql?(nil)
  edge.set_attribute "info", "initial_length", edge.length
 end
end


#assign masses by edge length
def vertex_mass(edge, density)
 mass=edge.start.get_attribute "info", "mass", 0
 length=edge.get_attribute "info", "initial_length"
 mass+=length*density/2
 edge.start.set_attribute "info", "mass", mass
 mass=edge.end.get_attribute "info", "mass", 0
 mass+=length*density/2
 edge.end.set_attribute "info", "mass", mass
end

#adds force to vertex
def add_force(vertex, force)
 zero=Geom::Vector3d.new 0,0,0
 force_sum=vertex.get_attribute "info", "force_sum", zero
 force_sum=force_sum+force
 vertex.set_attribute "info", "force_sum", force_sum
end


#counts stress for selected edge and assigns forces to ends
def edge_stress(edge, k)
 l0=edge.get_attribute "info", "initial_length"
 l=edge.length
 edge_start=edge.start
 edge_end=edge.end
 stress=(l/l0-1)*k
 force_start=edge.line[1]
 force_end=edge.line[1]
 force_start.length=stress
 force_end.length=-stress
 add_force(edge_start, force_start)
 add_force(edge_end, force_end)
 stress
end

#moves vertices (not physicaly accurate)
def move_vertices(entities, vertices, time)
 vectors=[]
 zero=Geom::Vector3d.new 0,0,0
 vertices.each { |vertex|
  type=vertex.get_attribute "info", "type" 
  if type=="pin"
   force_sum=zero
  elsif type=="clone" 
   other=vertex.edges[0].other_vertex vertex
   force_sum=other.get_attribute "info", "force_sum", zero
   mass=other.get_attribute "info", "mass", 1
  elsif type=="pinclone"
   other=vertex.edges[0].other_vertex vertex
   force_sum=zero-(other.get_attribute "info", "force_sum", zero)
   mass=vertex.get_attribute "info", "mass", 1
  else
   force_sum=vertex.get_attribute "info", "force_sum", zero
   mass=vertex.get_attribute "info", "mass", 1
  end
  if force_sum!=zero
   vector=force_sum
   vector.length=(force_sum.length*mass*time)
  else
   vector=zero
  end
  vectors.push vector
 }
 entities.transform_by_vectors(vertices, vectors)
end

def reset_forces(vertices)
 zero=Geom::Vector3d.new 0,0,0
 vertices.each { |vertex|
  vertex.set_attribute "info", "force_sum", zero
 }
end

#tags edges..

#tags vertices (helper)..

#colors edges
def edge_paint(edge,stress)
 s=25
 r=(-stress*s*5).round.to_f/5
 g=((1-stress.abs*s)*5).round.to_f/5
 b=(stress*s*5).round.to_f/5
 edge.material=[r, g, b]
end

def simple_springs_start
 #INITIALIZATION  
 
 #stop animation if running
 if defined? $timer_id
  UI.stop_timer $timer_id
 end

 #make entities array
 model = Sketchup.active_model
 entities = model.entities

 #list edges and vertices
 edges=list_edges(entities)
 vertices=list_vertices(entities)

 #rememeber lengths and assign masses
 edges.each { |edge|
  remember_length(edge)
  vertex_mass(edge, 1)
 }
 
 #reset forces at vertices
 reset_forces(vertices)

 #show dialog
 pola = ["Color stress:"]
 domyslnie = ["Yes"]
 lista = ["Yes|No"]
 query=UI.inputbox pola, domyslnie, lista, "Simple Springs"
 display=query[0]

 #turn off UI refresh
 #~status = model.start_operation "Spring animation", true

 #change edge display to "by material"
 Sketchup.active_model.rendering_options["EdgeColorMode"]=0 if display=="Yes"
 Sketchup.active_model.rendering_options["EdgeColorMode"]=1 if display=="No"

 #APPROXIMATION
 
 t=0.02
 k=1
 i=0
 $timer_id = UI.start_timer(t, true) {
  #count stress and assign forces to ends
  edges.each { |edge|
   stress=edge_stress(edge,1)
   edge_paint(edge,stress) if (display=="Yes")
  }
  #move vertices
  move_vertices(entities,vertices,t)
  reset_forces(vertices)
  #export view to jpg file
  #~file="spring"+i.to_s+".jpg"
  #~model.active_view.write_image file, 400, 250, false, 0.8
  #~i+=1
 }

end

def simple_springs_stop
 if defined? $timer_id
  UI.stop_timer $timer_id
 end
 Sketchup.active_model.rendering_options["EdgeColorMode"]=1
end

#INTERFACE

#add items to "Plugins" menu
menu = UI.menu "Plugins"
menu.add_separator
menu.add_item("Simple Springs - Start") {
 simple_springs_start
}
menu.add_item("Simple Springs - Stop") {
 simple_springs_stop
}

#add items to edge context menu
UI.add_context_menu_handler { |menu|
 edge=Sketchup.active_model.selection[0]
 if (edge.kind_of? Sketchup::Edge)
  menu.add_separator
  menu.add_item("Edge to Force") {
   if edge.start.edges.length == 1
    vertex=edge.start
   else
    vertex=edge.end
   end
   vertex.set_attribute "info", "type", "clone"
   edge.set_attribute "info", "initial_length", edge.length
  }
  menu.add_item("Edge to Pin") {
   if edge.start.edges.length == 1
    vertex=edge.start
    other=edge.end
   else
    vertex=edge.end
    other=edge.start
   end
   other.set_attribute "info", "type", "pin"
   vertex.set_attribute "info", "type", "pinclone"
  }
 end
}
